local _G = _G
local date = _G.date
local MAJOR_VERSION = 'GraveLevelBar'
local MINOR_VERSION = type('20110228221947') == 'number' and '20110228221947' or date('%Y%m%d%H%M%S')
local MODULE_NAME = 'xp'

local LibStub = _G.LibStub
local AceAddon = LibStub('AceAddon-3.0')
local addon = AceAddon:GetAddon(MAJOR_VERSION)

local module = addon:createModule(MODULE_NAME, 'AceEvent-3.0')

local UnitXP = _G.UnitXP
local UnitXPMax = _G.UnitXPMax
local UnitLevel = _G.UnitLevel
local GetMaxPlayerLevel = _G.GetMaxPlayerLevel
local GetXPExhaustion = _G.GetXPExhaustion
local min = min

module.defaults = addon:merge(module.defaults, {
    startXP = 0,
    targetXP = nil,
})

module.optionsTable = addon:merge(module.optionsTable, {
    name = "XP Bar",
    type = 'group',
    args = {
        startXP = {
            name = "Start XP",
            desc = "Minimum XP value to display.",
            type = 'range',
            min = 0,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getStartXP',
            set = 'setStartXP',
        },
        startLevel = {
            name = "Start Level",
            desc = "Minimum level to display.",
            type = 'range',
            min = 1,
            max = nil,
            bigStep = 1,
            get = 'getStartLevel',
            set = 'setStartLevel',
        },
        targetXP = {
            name = "Target XP",
            desc = "Maximum XP value to disply.",
            type = 'range',
            min = nil,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getTargetXP',
            set = 'setTargetXP',
        },
        targetLevel = {
            name = "Target Level",
            desc = "Maximum level to disply.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            get = 'getTargetLevel',
            set = 'setTargetLevel',
        },
    },
})

-- start xp
function module:getStartXP(info)
    return self:getOptions().startXP
end

function module:getStartXPMax()
    return self:getCurrentXP()
end

function module:setStartXP(info,value)
    self:getOptions().startXP = value
    self:Invalidate()
end

-- start level
function module:getStartLevel(info)
    local startXP = self:getStartXP()
    local level = self:calcLevelForXP(startXP)
    return level
end

function module:getStartLevelMax()
    local startXPMax = self:getStartXPMax()
    local levelMax = self:calcLevelForXP(startXPMax)
    return levelMax
end

function module:setStartLevel(info,value)
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local startXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setStartXP(info,startXP)
    else
        self:setStartXP(info,prevLevelXP)
    end
end

-- target xp
function module:getTargetXP(info)
    return self:getOptions().targetXP or self:getTargetXPMax()
end

function module:getTargetXPMax()
    local maxPlayerLevel = self:getTargetLevelMax()
    return self:calcXpForLevel(maxPlayerLevel)
end

function module:setTargetXP(info,value)
    self:getOptions().targetXP = value
    self:Invalidate()
end

-- target level
function module:getTargetLevel(info)
    local targetXP = self:getTargetXP()
    local level = self:calcLevelForXP(targetXP)
    return level
end

function module:getTargetLevelMax()
    return GetMaxPlayerLevel('player')
end

function module:setTargetLevel(info,value)
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local targetXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setTargetXP(info,targetXP)
    else
        self:setTargetXP(info,prevLevelXP)
    end
end

module.oldInitialize = module.OnInitialize
function module:OnInitialize()
    self:oldInitialize()
end

module.oldEnable = module.OnEnable
function module:OnEnable()
    self:oldEnable()
    self:RegisterEvent('UPDATE_EXHAUSTION', 'Invalidate')
    self:RegisterEvent('PLAYER_XP_UPDATE', 'Invalidate')
end

function module:Redraw()
    if not self.invalid then
        return
    end
    self.invalid = nil

    --addon:Debug("Redraw")

    local xpOptions = self:getOptions()
    local barCache = self.cache

    local maxPlayerLevel = GetMaxPlayerLevel()
    if maxPlayerLevel == 0 then
        self:Invalidate()
        return
    end

    local playerLevel = UnitLevel('player')
    local currentGlobalXP = self:getCurrentXP()
    -- update max settings if necessary
    if currentGlobalXP > xpOptions.targetXP then
        xpOptions.targetXP = currentGlobalXP
    end
    local startGlobalXP = xpOptions.startXP
    local targetGlobalXP = xpOptions.targetXP or self:calcXpForLevel(maxPlayerLevel)
    local restedGlobalXP = GetXPExhaustion()
    local currentGlobalXPRatio = (currentGlobalXP-startGlobalXP)/(targetGlobalXP-startGlobalXP)
    local restedGlobalXPRatio =  (restedGlobalXP               )/(targetGlobalXP-startGlobalXP)
    -- create textures
    assert(barCache and barCache.frame)
    local frame = barCache.frame
    -- texture for curent xp
    barCache.currentXPTexture = barCache.currentXPTexture or frame:CreateTexture(nil,'ARTWORK')
    local currentXPTexture = barCache.currentXPTexture
    currentXPTexture:ClearAllPoints()
    currentXPTexture:SetPoint('TOPLEFT')
    currentXPTexture:SetPoint('BOTTOM')
    currentXPTexture:SetWidth(currentGlobalXPRatio * frame:GetWidth())
    currentXPTexture:SetTexture(0,0,1,1)

    -- texture for rested xp
    barCache.restedXPTexture = barCache.restedXPTexture or frame:CreateTexture(nil,'ARTWORK')
    local restedXPTexture = barCache.restedXPTexture
    restedXPTexture:ClearAllPoints()
    restedXPTexture:SetPoint('TOP')
    restedXPTexture:SetPoint('BOTTOM')
    restedXPTexture:SetPoint('LEFT', currentXPTexture, 'RIGHT')
    local restedWidth = min(restedGlobalXPRatio,1-currentGlobalXPRatio) * frame:GetWidth()
    restedXPTexture:SetWidth(restedWidth)
    restedXPTexture:SetTexture(1,0,1,1)

    -- textures and numbers for levels
    barCache.levelTextures = barCache.levelTextures or {}
    barCache.levelStrings = barCache.levelStrings or {}
    local levelTextures = barCache.levelTextures
    local levelStrings = barCache.levelStrings
    for level = 1,maxPlayerLevel do
        local xpForLeveL = self:calcXpForLevel(level)
        if startGlobalXP <= xpForLeveL and xpForLeveL <= targetGlobalXP then
            local levelGlobalXPRatio = (xpForLeveL-startGlobalXP)/(targetGlobalXP-startGlobalXP)
            -- textures
            levelTextures[level] = levelTextures[level] or frame:CreateTexture(frame:GetName()..'Level'..level,'OVERLAY')
            local levelTexture = levelTextures[level]
            levelTexture:ClearAllPoints()
            levelTexture:SetPoint('TOP')
            levelTexture:SetPoint('BOTTOM')
            if levelGlobalXPRatio < 1.0 then
                levelTexture:SetPoint('LEFT',frame,'LEFT',levelGlobalXPRatio * frame:GetWidth(),0)
            else
                levelTexture:SetPoint('RIGHT')
            end
            levelTexture:SetWidth(1)
            levelTexture:SetTexture(1,1,1,1)
            levelTexture:Show()
            -- numbers
            levelStrings[level] = levelStrings[level] or frame:CreateFontString(nil,OVERLAY)
            local levelString = levelStrings[level]
            local isValid = levelString:SetFont('Fonts\\FRIZQT__.TTF',xpOptions.size)
            assert(isValid)
            levelString:ClearAllPoints()
            if levelGlobalXPRatio > 0 then
                levelString:SetPoint('RIGHT',levelTexture,'LEFT')
                levelString:SetJustifyH('RIGHT')
            else
                levelString:SetPoint('LEFT',levelTexture,'RIGHT')
                levelString:SetJustifyH('LEFT')
            end
            levelString:SetJustifyV('MIDDLE')
            levelString:SetText(level)
            levelString:Show()
        else
            if levelTextures[level] then
                levelTextures[level]:Hide()
            end
            if levelStrings[level] then
                levelStrings[level]:Hide()
            end
        end
    end

    --update options table
    self.optionsTable.args.startXP.max = self:getStartXPMax()
    self.optionsTable.args.startLevel.max = self:getStartLevelMax()
    self.optionsTable.args.targetXP.max = self:getTargetXPMax()
    self.optionsTable.args.targetLevel.max = self:getTargetLevelMax()
    self.optionsTable.args.targetXP.min = self.optionsTable.args.startXP.max
    self.optionsTable.args.targetLevel.min = self.optionsTable.args.startLevel.max
end

function module:getCurrentXP()
    local xpInCurrentLevel = UnitXP('player')
    local level = UnitLevel('player')
    return xpInCurrentLevel + self:calcXpForLevel(level)
end

function module:calcXpForLevel(level)
    local sum = 0
    for i=1,level-1 do
        sum = sum + XP_PER_LEVEL[i]
    end
    return sum
end

function module:calcLevelForXP(xp)
    local level = 1
    while xp > self:calcXpForLevel(level+1) do
        level = level + 1
    end
    local startXP = self:calcXpForLevel(level)
    assert(startXP <= xp,"level=" .. level .. " ,xp=" .. xp .. " ,startXP=" .. startXP)
    if startXP == xp then
        return level
    end
    local nextXP = self:calcXpForLevel(level+1)
    local fraction = (xp-startXP)/(nextXP-startXP)
    assert(0.0 <= fraction and fraction <= 1.0, "fraction=" .. fraction .. ", level=" .. level .. " ,xp=" .. xp .. " ,startXP=" .. startXP)
    return level + fraction
end

